from django.conf.urls import url
from chtest import views

urlpatterns = [url(r'^$', views.index)]

try:
    from chtest import consumers

    channel_routing = {"websocket.receive": consumers.ws_message}
except:
    pass



# Build paths inside the project like this: os.path.join(BASE_DIR, ...)

import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

SECRET_KEY = '-3yt98bfvxe)7+^h#(@8k#1(1m_fpd9x3q2wolfbf^!r5ma62u'

DEBUG = True

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
)

ROOT_URLCONF = 'testproject.urls'

WSGI_APPLICATION = 'testproject.wsgi.application'

STATIC_URL = "/static/"

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
        'TEST': {
            'NAME': os.path.join(BASE_DIR, 'test_db.sqlite3'),
        },
    },
}

ALLOWED_HOSTS = ['*']



# Settings for channels specifically
from testproject.settings.base import *

INSTALLED_APPS += (
    'channels',
)

CHANNEL_LAYERS = {
    "default": {
        "BACKEND": "asgi_ipc.IPCChannelLayer",
        "ROUTING": "testproject.urls.channel_routing",
    },
}



from locust import HttpLocust, TaskSet, task

class UserBehavior(TaskSet):
    def on_start(self):
        """ on_start is called when a Locust start before any task is scheduled """
        self.index()

    @task
    def index(self):
        self.client.get("/")


class WebsiteUser(HttpLocust):
    task_set = UserBehavior
    min_wait=5000
    max_wait=9000



from channels import DEFAULT_CHANNEL_LAYER
from channels.asgi import InvalidChannelLayerError, channel_layers
from channels.test import ChannelTestCase
from django.test import override_settings


class TestChannelLayerManager(ChannelTestCase):

    def test_config_error(self):
        """
        If channel layer doesn't specify TEST_CONFIG, `make_test_backend`
        should result into error.
        """

        with self.assertRaises(InvalidChannelLayerError):
            channel_layers.make_test_backend(DEFAULT_CHANNEL_LAYER)

    @override_settings(CHANNEL_LAYERS={
        'default': {
            'BACKEND': 'asgiref.inmemory.ChannelLayer',
            'ROUTING': [],
            'TEST_CONFIG': {
                'expiry': 100500,
            },
        },
    })
    def test_config_instance(self):
        """
        If channel layer provides TEST_CONFIG, `make_test_backend` should
        return channel layer instance appropriate for testing.
        """

        layer = channel_layers.make_test_backend(DEFAULT_CHANNEL_LAYER)
        self.assertEqual(layer.channel_layer.expiry, 100500)



# Settings for channels specifically
from testproject.settings.base import *

INSTALLED_APPS += (
    'channels',
)

CHANNEL_LAYERS = {
    "default": {
        "BACKEND": "asgi_redis.RedisChannelLayer",
        "ROUTING": "testproject.urls.channel_routing",
        "CONFIG": {
            "hosts": [os.environ.get('REDIS_URL', 'redis://127.0.0.1:6379')],
        },
        "TEST_CONFIG": {
            "hosts": [os.environ.get('REDIS_URL', 'redis://127.0.0.1:6379')],
        },
    },
}



from fabric.api import sudo, task, cd

# CHANNEL TASKS
@task
def setup_redis():
    sudo("apt-get update && apt-get install -y redis-server")
    sudo("sed -i -e 's/127.0.0.1/0.0.0.0/g' /etc/redis/redis.conf")
    sudo("/etc/init.d/redis-server stop")
    sudo("/etc/init.d/redis-server start")


@task
def setup_channels():
    sudo("apt-get update && apt-get install -y git python-dev python-setuptools python-pip")
    sudo("pip install -U pip")
    sudo("pip install -U asgi_redis asgi_ipc git+https://github.com/django/daphne.git@#egg=daphne")
    sudo("rm -rf /srv/channels")
    sudo("git clone https://github.com/django/channels.git /srv/channels/")
    with cd("/srv/channels/"):
        sudo("python setup.py install")


@task
def run_daphne(redis_ip):
    with cd("/srv/channels/testproject/"):
        sudo("REDIS_URL=redis://%s:6379 daphne -b 0.0.0.0 -p 80 testproject.asgi.redis:channel_layer" % redis_ip)


@task
def run_worker(redis_ip):
    with cd("/srv/channels/testproject/"):
        sudo("REDIS_URL=redis://%s:6379 python manage.py runworker" % redis_ip)


# Current loadtesting setup
@task
def setup_load_tester(src="https://github.com/django/channels.git"):
    sudo("apt-get update && apt-get install -y git nodejs && apt-get install npm")
    sudo("npm install -g loadtest")
    sudo("ln -s /usr/bin/nodejs /usr/bin/node")


# Run current loadtesting setup
# example usage: $ fab run_loadtest:http://127.0.0.1,rps=10 -i "id_rsa" -H ubuntu@example.com
@task
def run_loadtest(host, t=90):
    sudo("loadtest -c 10 -t {t} {h}".format(h=host, t=t))

# Run current loadtesting setup
# example usage: $ fab run_loadtest:http://127.0.0.1,rps=10 -i "id_rsa" -H ubuntu@example.com
@task
def run_loadtest_rps(host, t=90, rps=200):
    sudo("loadtest -c 10 --rps {rps} -t {t} {h}".format(h=host, t=t, rps=rps))


# Task that Andrew used for loadtesting earlier on
@task
def setup_tester():
    sudo("apt-get update && apt-get install -y apache2-utils python3-pip")
    sudo("pip3 -U pip autobahn twisted")
    sudo("rm -rf /srv/channels")
    sudo("git clone https://github.com/django/channels.git /srv/channels/")


@task
def shell():
    sudo("bash")



from __future__ import unicode_literals

from django.apps import apps
from django.conf import settings
from django.core.management import BaseCommand, CommandError

from channels import DEFAULT_CHANNEL_LAYER, channel_layers
from channels.log import setup_logger
from channels.signals import worker_process_ready
from channels.staticfiles import StaticFilesConsumer
from channels.worker import Worker, WorkerGroup


class Command(BaseCommand):

    leave_locale_alone = True

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser.add_argument(
            '--layer', action='store', dest='layer', default=DEFAULT_CHANNEL_LAYER,
            help='Channel layer alias to use, if not the default.',
        )
        parser.add_argument(
            '--only-channels', action='append', dest='only_channels',
            help='Limits this worker to only listening on the provided channels (supports globbing).',
        )
        parser.add_argument(
            '--exclude-channels', action='append', dest='exclude_channels',
            help='Prevents this worker from listening on the provided channels (supports globbing).',
        )
        parser.add_argument(
            '--threads', action='store', dest='threads',
            default=1, type=int,
            help='Number of threads to execute.'
        )
        parser.add_argument(
            '--disable-graceful-stop', action='store_false', dest='stop_gracefully',
            help='Disable graceful stop for a worker group (instant exit on termination signal).',
        )

    def handle(self, *args, **options):
        # Get the backend to use
        self.verbosity = options.get("verbosity", 1)
        self.logger = setup_logger('django.channels', self.verbosity)
        self.channel_layer = channel_layers[options.get("layer", DEFAULT_CHANNEL_LAYER)]
        self.n_threads = options.get('threads', 1)
        # Check that handler isn't inmemory
        if self.channel_layer.local_only():
            raise CommandError(
                "You cannot span multiple processes with the in-memory layer. " +
                "Change your settings to use a cross-process channel layer."
            )
        # Check a handler is registered for http reqs
        # Serve static files if Django in debug mode
        if settings.DEBUG and apps.is_installed('django.contrib.staticfiles'):
            self.channel_layer.router.check_default(http_consumer=StaticFilesConsumer())
        else:
            self.channel_layer.router.check_default()
        # Optionally provide an output callback
        callback = None
        if self.verbosity > 1:
            callback = self.consumer_called
        self.callback = callback
        self.options = options
        # Choose an appropriate worker.
        worker_kwargs = {}
        if self.n_threads == 1:
            self.logger.info("Using single-threaded worker.")
            worker_cls = Worker
        else:
            self.logger.info("Using multi-threaded worker, {} thread(s).".format(self.n_threads))
            worker_cls = WorkerGroup
            worker_kwargs['n_threads'] = self.n_threads
            worker_kwargs['stop_gracefully'] = self.options.get('stop_gracefully')
        # Run the worker
        self.logger.info("Running worker against channel layer %s", self.channel_layer)
        try:
            worker = worker_cls(
                channel_layer=self.channel_layer,
                callback=self.callback,
                only_channels=self.options.get("only_channels", None),
                exclude_channels=self.options.get("exclude_channels", None),
                **worker_kwargs
            )
            worker_process_ready.send(sender=worker)
            worker.ready()
            worker.run()
        except KeyboardInterrupt:
            pass

    def consumer_called(self, channel, message):
        self.logger.debug("%s", channel)



from __future__ import unicode_literals

import time
import random
import statistics
from autobahn.twisted.websocket import (
    WebSocketClientProtocol,
    WebSocketClientFactory,
)
from twisted.internet import reactor

stats = {}


class MyClientProtocol(WebSocketClientProtocol):

    def __init__(self, *args, **kwargs):
        WebSocketClientProtocol.__init__(self, *args, **kwargs)
        self.fingerprint = "".join(random.choice("abcdefghijklmnopqrstuvwxyz") for i in range(16))
        stats[self.fingerprint] = {}

    def onConnect(self, response):
        self.opened = time.time()
        self.sent = 0
        self.last_send = None
        self.received = 0
        self.corrupted = 0
        self.out_of_order = 0
        self.latencies = []

    def onOpen(self):
        def hello():
            if self.last_send is None:
                if self.sent >= self.factory.num_messages:
                    self.sendClose()
                    return
                self.last_send = time.time()
                self.sendMessage(("%s:%s" % (self.sent, self.fingerprint)).encode("ascii"))
                self.sent += 1
            else:
                # Wait for receipt of ping
                pass
            self.factory.reactor.callLater(1.0 / self.factory.message_rate, hello)
        hello()

    def onMessage(self, payload, isBinary):
        # Detect receive-before-send
        if self.last_send is None:
            self.corrupted += 1
            print("CRITICAL: Socket %s received before sending: %s" % (self.fingerprint, payload))
            return
        num, fingerprint = payload.decode("ascii").split(":")
        if fingerprint != self.fingerprint:
            self.corrupted += 1
        try:
            if int(num) != self.received:
                self.out_of_order += 1
        except ValueError:
            self.corrupted += 1
        self.latencies.append(time.time() - self.last_send)
        self.received += 1
        self.last_send = None

    def onClose(self, wasClean, code, reason):
        if hasattr(self, "sent"):
            stats[self.fingerprint] = {
                "sent": self.sent,
                "received": self.received,
                "corrupted": self.corrupted,
                "out_of_order": self.out_of_order,
                "latencies": self.latencies,
                "connect": True,
            }
        else:
            stats[self.fingerprint] = {
                "sent": 0,
                "received": 0,
                "corrupted": 0,
                "out_of_order": 0,
                "connect": False,
            }



class Benchmarker(object):
    """
    Performs benchmarks against WebSockets.
    """

    def __init__(self, url, num, concurrency, rate, messages, spawn):
        self.url = url
        self.num = num
        self.concurrency = concurrency
        self.rate = rate
        self.spawn = spawn
        self.messages = messages
        self.factory = WebSocketClientFactory(self.url)
        self.factory.protocol = MyClientProtocol
        self.factory.num_messages = self.messages
        self.factory.message_rate = self.rate

    def loop(self):
        self.spawn_loop()
        self.progress_loop()

    def spawn_loop(self):
        self.spawn_connections()
        reactor.callLater(0.1, self.spawn_loop)

    def progress_loop(self):
        self.print_progress()
        reactor.callLater(1, self.progress_loop)

    def spawn_connections(self):
        # Stop spawning if we did the right total number
        max_to_spawn = self.num - len(stats)
        if max_to_spawn <= 0:
            return
        # Don't spawn too many at once
        max_to_spawn = min(max_to_spawn, int(self.spawn / 10.0))
        # Decode connection args
        host, port = self.url.split("://")[1].split(":")
        port = int(port)
        # Only spawn enough to get up to concurrency
        open_protocols = len([x for x in stats.values() if not x])
        to_spawn = min(max(self.concurrency - open_protocols, 0), max_to_spawn)
        for _ in range(to_spawn):
            reactor.connectTCP(host, port, self.factory)

    def print_progress(self):
        open_protocols = len([x for x in stats.values() if not x])
        print("%s open, %s total" % (
            open_protocols,
            len(stats),
        ))
        if open_protocols == 0 and len(stats) >= self.num:
            reactor.stop()
            self.print_stats()

    def percentile(self, values, fraction):
        """
        Returns a percentile value (e.g. fraction = 0.95 -> 95th percentile)
        """
        values = sorted(values)
        stopat = int(len(values) * fraction)
        if stopat == len(values):
            stopat -= 1
        return values[stopat]

    def print_stats(self):
        # Collect stats together
        latencies = []
        num_good = 0
        num_incomplete = 0
        num_failed = 0
        num_corruption = 0
        num_out_of_order = 0
        for entry in stats.values():
            latencies.extend(entry.get("latencies", []))
            if not entry['connect']:
                num_failed += 1
            elif entry['sent'] != entry['received']:
                num_incomplete += 1
            elif entry['corrupted']:
                num_corruption += 1
            elif entry['out_of_order']:
                num_out_of_order += 1
            else:
                num_good += 1

        if latencies:
            # Some analysis on latencies
            latency_mean = statistics.mean(latencies)
            latency_median = statistics.median(latencies)
            latency_stdev = statistics.stdev(latencies)
            latency_95 = self.percentile(latencies, 0.95)
            latency_99 = self.percentile(latencies, 0.99)

        # Print results
        print("-------")
        print("Sockets opened: %s" % len(stats))
        if latencies:
            print("Latency stats: Mean %.3fs  Median %.3fs  Stdev %.3f  95%% %.3fs  99%% %.3fs" % (
                latency_mean,
                latency_median,
                latency_stdev,
                latency_95,
                latency_99,
            ))
        print("Good sockets: %s (%.2f%%)" % (num_good, (float(num_good) / len(stats))*100))
        print("Incomplete sockets: %s (%.2f%%)" % (num_incomplete, (float(num_incomplete) / len(stats))*100))
        print("Corrupt sockets: %s (%.2f%%)" % (num_corruption, (float(num_corruption) / len(stats))*100))
        print("Out of order sockets: %s (%.2f%%)" % (num_out_of_order, (float(num_out_of_order) / len(stats))*100))
        print("Failed to connect: %s (%.2f%%)" % (num_failed, (float(num_failed) / len(stats))*100))


if __name__ == '__main__':

    import sys
    import argparse

    from twisted.python import log

#    log.startLogging(sys.stdout)

    parser = argparse.ArgumentParser()
    parser.add_argument("url")
    parser.add_argument("-n", "--num", type=int, default=100, help="Total number of sockets to open")
    parser.add_argument("-c", "--concurrency", type=int, default=10, help="Number of sockets to open at once")
    parser.add_argument("-r", "--rate", type=float, default=1, help="Number of messages to send per socket per second")
    parser.add_argument("-m", "--messages", type=int, default=5, help="Number of messages to send per socket before close")
    parser.add_argument("-s", "--spawn", type=int, default=30, help="Number of sockets to spawn per second, max")
    args = parser.parse_args()

    benchmarker = Benchmarker(
        url=args.url,
        num=args.num,
        concurrency=args.concurrency,
        rate=args.rate,
        messages=args.messages,
        spawn=args.spawn,
    )
    benchmarker.loop()
    reactor.run()



# Settings for channels specifically
from testproject.settings.base import *

INSTALLED_APPS += (
    'channels',
)

CHANNEL_LAYERS = {
    'default': {
        'BACKEND': 'asgi_rabbitmq.RabbitmqChannelLayer',
        'ROUTING': 'testproject.urls.channel_routing',
        'CONFIG': {
            'url': os.environ.get(
                'RABBITMQ_URL',
                'amqp://guest:guest@localhost:5672/%2F',
            ),
        },
    },
}



from setuptools import find_packages, setup

setup(
    name='channels-benchmark',
    packages=find_packages(),
    py_modules=['benchmark'],
    install_requires=[
        'autobahn',
        'Twisted',
        'statistics ; python_version < "3.0"',
    ],
)



from __future__ import unicode_literals

from django.http.cookie import parse_cookie

from channels import route
from channels.exceptions import ChannelSocketException
from channels.handler import AsgiRequest
from channels.test import ChannelTestCase, WSClient, apply_routes
from channels.sessions import enforce_ordering


class WSClientTests(ChannelTestCase):
    def test_cookies(self):
        client = WSClient()
        client.set_cookie('foo', 'not-bar')
        client.set_cookie('foo', 'bar')
        client.set_cookie('qux', 'qu;x')

        # Django's interpretation of the serialized cookie.
        cookie_dict = parse_cookie(client.headers['cookie'].decode('ascii'))

        self.assertEqual(client.get_cookies(),
                         cookie_dict)

        self.assertEqual({'foo': 'bar',
                          'qux': 'qu;x',
                          'sessionid': client.get_cookies()['sessionid']},
                         cookie_dict)

    def test_simple_content(self):
        client = WSClient()
        content = client._get_content(text={'key': 'value'}, path='/my/path')

        self.assertEqual(content['text'], '{"key": "value"}')
        self.assertEqual(content['path'], '/my/path')
        self.assertTrue('reply_channel' in content)
        self.assertTrue('headers' in content)

    def test_path_in_content(self):
        client = WSClient()
        content = client._get_content(content={'path': '/my_path'}, text={'path': 'hi'}, path='/my/path')

        self.assertEqual(content['text'], '{"path": "hi"}')
        self.assertEqual(content['path'], '/my_path')
        self.assertTrue('reply_channel' in content)
        self.assertTrue('headers' in content)

    def test_session_in_headers(self):
        client = WSClient()
        content = client._get_content()
        self.assertTrue('path' in content)
        self.assertEqual(content['path'], '/')

        self.assertTrue('headers' in content)
        self.assertIn(b'cookie', [x[0] for x in content['headers']])
        self.assertIn(b'sessionid', [x[1] for x in content['headers'] if x[0] == b'cookie'][0])

    def test_ordering_in_content(self):
        client = WSClient(ordered=True)
        content = client._get_content()
        self.assertTrue('order' in content)
        self.assertEqual(content['order'], 0)
        client.order = 2
        content = client._get_content()
        self.assertTrue('order' in content)
        self.assertEqual(content['order'], 2)

    def test_ordering(self):

        client = WSClient(ordered=True)

        @enforce_ordering
        def consumer(message):
            message.reply_channel.send({'text': message['text']})

        with apply_routes(route('websocket.receive', consumer)):
            client.send_and_consume('websocket.receive', text='1')  # order = 0
            client.send_and_consume('websocket.receive', text='2')  # order = 1
            client.send_and_consume('websocket.receive', text='3')  # order = 2

            self.assertEqual(client.receive(), 1)
            self.assertEqual(client.receive(), 2)
            self.assertEqual(client.receive(), 3)

    def test_get_params(self):
        client = WSClient()
        content = client._get_content(path='/my/path?test=1&token=2')
        self.assertTrue('path' in content)
        self.assertTrue('query_string' in content)
        self.assertEqual(content['path'], '/my/path')
        self.assertEqual(content['query_string'], 'test=1&token=2')

    def test_get_params_with_consumer(self):
        client = WSClient(ordered=True)

        def consumer(message):
            message.content['method'] = 'FAKE'
            message.reply_channel.send({'text': dict(AsgiRequest(message).GET)})

        with apply_routes([route('websocket.receive', consumer, path=r'^/test'),
                           route('websocket.connect', consumer, path=r'^/test')]):
            path = '/test?key1=val1&key2=val2&key1=val3'
            client.send_and_consume('websocket.connect', path=path, check_accept=False)
            self.assertDictEqual(client.receive(), {'key2': ['val2'], 'key1': ['val1', 'val3']})

            client.send_and_consume('websocket.receive', path=path)
            self.assertDictEqual(client.receive(), {})

    def test_channel_socket_exception(self):

        class MyChannelSocketException(ChannelSocketException):

            def run(self, message):
                message.reply_channel.send({'text': 'error'})

        def consumer(message):
            raise MyChannelSocketException

        client = WSClient()
        with apply_routes(route('websocket.receive', consumer)):
            client.send_and_consume('websocket.receive')

            self.assertEqual(client.receive(json=False), 'error')



from __future__ import unicode_literals

from django.test import override_settings
from channels.exceptions import DenyConnection
from channels.security.websockets import allowed_hosts_only
from channels.message import Message
from channels.test import ChannelTestCase


@allowed_hosts_only
def connect(message):
    return True


class OriginValidationTestCase(ChannelTestCase):

    @override_settings(ALLOWED_HOSTS=['example.com'])
    def test_valid_origin(self):
        content = {
            'headers': [[b'origin', b'http://example.com']]
        }
        message = Message(content, 'websocket.connect', None)
        self.assertTrue(connect(message))

    @override_settings(ALLOWED_HOSTS=['example.com'])
    def test_invalid_origin(self):
        content = {
            'headers': [[b'origin', b'http://example.org']]
        }
        message = Message(content, 'websocket.connect', None)
        self.assertRaises(DenyConnection, connect, message)

    def test_invalid_origin_header(self):
        invalid_headers = [
            [],  # origin header missing
            [b'origin', b''],  # origin header empty
            [b'origin', b'\xc3\xa4']  # non-ascii
        ]
        for headers in invalid_headers:
            content = {
                'headers': [headers]
            }
            message = Message(content, 'websocket.connect', None)
            self.assertRaises(DenyConnection, connect, message)



from uuid import uuid4

from django.db import models


class TestUUIDModel(models.Model):
    """
    Simple model with UUIDField as primary key for tests.
    """

    id = models.UUIDField(primary_key=True, default=uuid4)
    name = models.CharField(max_length=255)
